def apply_template(sql: str, user_ns) -> str:
    try:
        from jinja2 import Template
    except Exception as e:
        raise ValueError(
            "Jinja2 must be installed to use --jinja: %pip install Jinja2"
        ) from e

    t = Template(sql)
    return t.render(user_ns)



import logging
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

# An experiment at showing the AST using SQLParse.

logger = logging.getLogger("magic_duckdb")

# Extra path is optional: you should have "dot" in your PATH. If not, you can set extra_path to the
# fully qualified path to your dot executable.
dot_path = None
# "c:\\Program files\\graphviz\\bin\\dot.exe"

# graphviz.backend.dot_command.DOT_BINARY = pathlib.Path("c:\\Program files\\graphviz\\bin\\dot")  # type: ignore # noqa
SUPPRESS_EMPTY = True
basic_types = (int, float, str, bool, complex)


@dataclass
class Node:
    id: int
    name: str
    parent: "Node"
    properties: Dict
    children: List["Node"]

    def props_str(self):
        return " ".join([f"{v}" for k, v in self.properties.items()])


def get_tree(ast_json) -> Tuple[List[Node], List[Tuple[Node, Node]]]:
    nodes: List[Node] = []
    edges: List[Tuple[Node, Node]] = []

    def get_node(name, parent):
        id = len(nodes)
        node = Node(id=id, name=name, parent=parent, properties={}, children=[])
        nodes.append(node)

        if parent is not None:
            parent.children.append(node)
            edges.append((parent, node))
        return node

    def process_node(name, o, parent: Optional[Node], use_parent: bool = False):
        if SUPPRESS_EMPTY:
            if o is None:
                return
            if isinstance(o, str) and o == "":
                return
            if isinstance(o, list) and len(o) == 0:
                return

        if use_parent:
            if parent is None:
                raise ValueError("Unexpected None parent")
            node = parent
        else:
            node = get_node(name, parent)
        if isinstance(o, basic_types):
            prop = node.properties.get("value")
            if prop is not None:
                # Only case where we hit this was column_names
                # Might need to change if there are other similar
                # cases where "." isn't the right delimiter
                node.properties["value"] = f"{prop}.{o}"
            else:
                node.properties["value"] = f"{o}"

        elif isinstance(o, dict):
            if "type" in o:
                node.properties["type"] = o["type"]
            if "class" in o and o["class"] != o.get("type"):
                node.properties["class"] = o["class"]
            for k, v in o.items():
                if k != "type" and k != "class":
                    process_node(k, v, node)
        elif isinstance(o, list):
            for obj in o:  # skip over
                process_node(name, obj, node, True)

    process_node("Root", ast_json, None)

    return nodes, edges


def _print_node(n: Node, depth, lines):
    indent = "-" * depth

    lines.append(f"{indent} | {n.name}: {n.props_str()}")
    for c in n.children:
        _print_node(c, depth + 1, lines)


def ast_tree(ast_json):
    nodes, edges = get_tree(ast_json)

    root = nodes[0]

    lines = []
    _print_node(root, 0, lines)

    return "\n".join(lines)


def ast_draw_graphviz(ast_json):
    try:
        # Defer loading, since this is optional
        import graphviz  # type: ignore # noqa

        logger.debug("Graphviz Python module is available")
    except Exception:
        logger.debug("Graphviz Python module not available")
        return None
    dot = graphviz.Digraph()  # type: ignore # noqa
    dot.attr(rankdir="LR")

    nodes, edges = get_tree(ast_json)

    for node in nodes:
        if node.properties is not None and len(node.properties) > 0:
            props = "\\n".join([f"{v}" for k, v in node.properties.items()])
        else:
            props = None
        node_label = f"{node.name}\\n{props}" if props is not None else f"<{node.name}>"
        dot.node(f"{node.id}", node_label, shape="rectangle")

    for e in edges:
        dot.edge(f"{e[0].id}", f"{e[1].id}", color="red")

    return dot



import json
from typing import Dict, List, Optional, Tuple

import duckdb
from IPython.display import Markdown

from magic_duckdb.extras.ast_graphviz import ast_draw_graphviz, ast_tree
from magic_duckdb.extras.explain_analyze_graphviz import draw_graphviz


def execute_db(
    *, query: str, con: duckdb.DuckDBPyConnection, execute: bool = False, params=None
):
    """Simple wrapper to allow alternative implementations or wrappers to be inserted
    execute = False: returns a Relation object"""

    if execute or params is not None:
        return con.execute(query, parameters=params)
    else:
        return con.sql(query)


class DuckDbMode:
    """Lightweight wrapper to separate duckdb specific logic from the Magic.
    Goal here is to expose all of DuckDB's goodness, but also make it easy to extend or replace this.
    """

    export_functions: List[str] = [
        "df",
        "df_markdown",
        "markdown",
        "arrow",
        "pl",
        "describe",
        "show",
        "relation",
    ]

    explain_functions: List[str] = [
        "explain",
        "explain_analyze_tree",
        "explain_analyze_json",
        "explain_analyze_draw",
        "analyze",
        "ast_json",
        "ast_draw",
        "ast_tree",
    ]

    def default_connection(self) -> duckdb.DuckDBPyConnection:
        if isinstance(duckdb.default_connection, duckdb.DuckDBPyConnection):
            return duckdb.default_connection
        else:
            return (
                duckdb.default_connection()
            )  # https://github.com/duckdb/duckdb/pull/13442 changed from property to function

    def connect(self, conn_string: str) -> duckdb.DuckDBPyConnection:
        return duckdb.connect(conn_string)

    def explain_analyze(
        self,
        query_string: str,
        connection: duckdb.DuckDBPyConnection,
        export_function,
        explain_function,
    ):
        # query_tree

        if explain_function == "explain_analyze_tree" or explain_function == "analyze":
            execute_db(
                query="PRAGMA enable_profiling=query_tree", con=connection, execute=True
            )
            r = execute_db(query=query_string, con=connection, execute=False)
            t = r.explain(type="analyze")  # type: ignore
            print(t)  # noqa: T201
            return t
        elif explain_function == "explain_analyze_json":
            execute_db(
                query="PRAGMA enable_profiling=json", con=connection, execute=True
            )
            r = execute_db(query=query_string, con=connection, execute=False)
            j = r.explain(type="analyze")  # type: ignore
            return j
        elif explain_function == "explain_analyze_draw":
            execute_db(
                query="PRAGMA enable_profiling=json", con=connection, execute=True
            )
            r = execute_db(query=query_string, con=connection, execute=False)
            j = r.explain(type="analyze")  # type: ignore
            return draw_graphviz(j)
        elif explain_function == "explain":
            r = execute_db(query=query_string, con=connection, execute=False)
            j = r.explain()  # type: ignore
            print(j)  # noqa: T201
            return j
        elif explain_function.startswith("ast"):
            r = connection.execute(
                "select json_serialize_sql($1::varchar)", [query_string]
            )
            df = r.df()
            json_str = df.iat[0, 0]
            json_obj = json.loads(json_str)
            if explain_function == "ast_json":
                return json_obj
            elif explain_function == "ast_draw":
                return ast_draw_graphviz(json_obj)
            elif explain_function == "ast_tree":
                return ast_tree(json_obj)
            else:
                raise ValueError(f"Unexpected AST mode {explain_function}")

        else:
            raise ValueError(f"Unexpected explain mode {explain_function}")

    def execute(
        self,
        query_string: str,
        connection: Optional[duckdb.DuckDBPyConnection] = None,
        export_function: Optional[str] = None,
        explain_function: Optional[str] = None,
        params: Optional[List[object]] = None,
        export_kwargs: Dict[str, object] = None,
    ) -> Tuple[object, bool] | Markdown:

        if export_kwargs is None:
            export_kwargs = {}

        if connection is None:
            connection = self.default_connection()

        if export_function is None:
            export_function = self.export_functions[0]

        if explain_function is not None:
            return self.explain_analyze(
                query_string, connection, export_function, explain_function
            )
        else:
            try:
                if export_function in ["show", "describe", "relation"]:
                    # Result must be a relation, not a connection
                    execute = False
                else:
                    execute = True
                r = execute_db(
                    query=query_string, con=connection, params=params, execute=execute
                )
            except Exception as e:
                e.add_note(f"While executing {query_string=}")
                raise

            if r is None or ("relation" == export_function):
                return r
            else:
                if export_function == "df_markdown" or export_function == "markdown":
                    md = r.df().to_markdown(index=False)
                    mdd = Markdown(md)
                    return mdd

                else:
                    export_function = export_function
                    f = getattr(r, export_function)

                    return f(**export_kwargs)



from magic_duckdb._version import __version__

# This must be modified before magic is initialized, before load_ipython_extension
MAGIC_NAME = "dql"

# Disable autocompletion initialization by setting this to False before loading extension
ENABLE_AUTOCOMPLETE = True


def load_ipython_extension(ip):
    """Load the extension in IPython."""
    if ip is None:
        raise ValueError("No ipython found")

    if ENABLE_AUTOCOMPLETE:
        from .autocomplete.common import init_completers

        init_completers(ip)

    from .magic import DuckDbMagic  # type: ignore

    ip.register_magics(DuckDbMagic)


__all__ = [
    "__version__",
]



import logging


def log_to_file(logfile: str):
    logger = logging.getLogger("magic_duckdb")
    logger.setLevel(logging.DEBUG)

    file_handler = logging.FileHandler(logfile)
    file_handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)



# autocompletion for pre-iPython 8.6.0
# the main limitation here is
import logging

from magic_duckdb.autocomplete.common import (
    get_column_names,
    get_table_names,
    sql_expects_tablename,
)

logger = logging.getLogger("magic_duckdb")


def line_completer(ipython, event):
    # v1 (pre 8.6.0 iPython) only passes the current line, not the entire text. So, line magics are fine but cell magics aren't

    try:
        logger.debug(event)
        command = event.command
        if command != "%dql" and command != "%%dql":
            return

        line = event.line
        symbol = event.symbol
        # text_until_cursor = event.text_until_cursor

        if symbol == "" and line[-1] == " ":
            last_word = line.split(" ")[-2]
            logger.info(last_word)

            if last_word.upper() in sql_expects_tablename:
                return get_table_names(ipython)
        elif symbol[-1] == ".":
            tablename = symbol[:-1]

            columns = get_column_names(ipython, tablename)
            logger.debug("Using columns %s, %s", columns, type(columns))

            # To complete a word, the current symbol needs to be prefixed
            # To suggest the next word, just return the next word
            results = [f"{symbol}{col}" for col in columns]
            logger.info(results)
            return results

        logger.info(event)
    except Exception:
        logger.exception("Error: %s", event)


def init_completer(ipython):
    ipython.set_hook("complete_command", line_completer, re_key="[%]+dql.*")
